﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
VirtualLock v0.4 by MacGruber.
Use VR head and hand tracking for virtual self-bondage.

Version 0.4 - 2023-01-03
	Allowing position adjustment for the head-up display.

Version 0.3 - 2022-01-03
	Added info text.
	Updated MacGruber_Utils.cs.

Version 0.2 - 2020-10-11
	Improved breach behavior by auto-unlock after some seconds.

Version 0.1 - 2020-09-26
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using SimpleJSON;

namespace MacGruber
{
    public class VirtualLock : MVRScript
    {
		private Transform myCanvas;
		private Image myTimerPanel;
		private Image myMessagePanel;
		private Image myHeadPanel;
		private Image myHandPanelL;
		private Image myHandPanelR;
		private Text myTimerText;
		private Text myMessageText;
		private Text myHeadDistanceText;
		private Text myHeadAngleText;
		private Text myHandDistanceTextL;
		private Text myHandDistanceTextR;	
		private bool myGameIsDesktop = false;
		private bool myEngaged = false;
		private float myFadeClock = 0.0f;
		private float myAlpha = 0.0f;
		private float myCountdown = 0.0f;
		private float myTimer = 0.0f;
		private float myBreachTimer = 0.0f;
		private int myCountdownSeconds = -1;
		private int myTimerSeconds = -1;
		private int myBreachTimerSeconds = -1;
		
		private Transform myHeadControl = null;
		private Transform myHandControlL = null;
		private Transform myHandControlR = null;
		private Transform myFocusPoint = null;
		
		private Vector3 myHeadPosition;
		private Vector3 myHandPositionL;
		private Vector3 myHandPositionR;
		private int myHeadDistance = -1;
		private int myHandDistanceL = -1;
		private int myHandDistanceR = -1;
		private int myHeadAngle = -1;
		
		private string myBundleURL = null;
		
		private JSONStorableFloat myWarnRadius;
		private JSONStorableFloat myBreachRadius;
		private JSONStorableFloat myWarnAngle;
		private JSONStorableFloat myBreachAngle;
		private JSONStorableFloat myBreachDuration;
		private JSONStorableFloat myCountdownTime;
		private JSONStorableFloat myLockTime;
		private EventTrigger myEventInit;
		private EventTrigger myEventLock;
		private EventTrigger myEventUnlock;
		private EventTrigger myEventClear;
		private EventTrigger myEventWarn;
		private EventTrigger myEventBreach;
		
		private JSONStorableFloat myHeadupAngleVR;
		private JSONStorableFloat myHeadupAngleDesktop;
		
		private int myState = STATE_CLEAR;
		
		private const int STATE_CLEAR = 0;
		private const int STATE_WARN = 1;
		private const int STATE_BREACH = 2;		
		
		
		private const float HEADUP_DISTANCE = 0.6f;
		private static readonly string BASE_PATH = "Assets/SecretRoom/UI/";		
		private static readonly Color BASE_COLOR = new Color(100.0f/255.0f, 100.0f/255.0f, 100.0f/255.0f);
		private static readonly Color BREACH_COLOR = Color.red;
		private const float FADE_TIME = 8.0f;
		
		
        public override void Init()
		{
			SuperController sc = SuperController.singleton;
			
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>VirtualLock v0.4</b></size>\nUse VR head and hand tracking for virtual self-bondage.</color>",
				250.0f, true
			);
			
			myEventInit = new EventTrigger(this, "OnInit");
			myEventLock = new EventTrigger(this, "OnLock");
			myEventUnlock = new EventTrigger(this, "OnUnlock");
			myEventClear = new EventTrigger(this, "OnClear");
			myEventWarn = new EventTrigger(this, "OnWarn");
			myEventBreach = new EventTrigger(this, "OnBreach");
								
			myHeadControl = CameraTarget.centerTarget.transform;
			myHandControlL = sc.leftHand;
			myHandControlR = sc.rightHand;
			
			myCanvas = null;
			myBundleURL = Utils.GetPluginPath(this) + "/MacGruber_VirtualLock.assetbundle";
			Request request = new AssetLoader.AssetBundleFromFileRequest {path = myBundleURL, callback = OnAssetBundleLoaded};
			AssetLoader.QueueLoadAssetBundleFromFile(request);
			
			Utils.SetupButton(this, "Init", InitLock, false);
			Utils.SetupButton(this, "Abort", AbortLock, false);
			Utils.SetupAction(this, "Init", InitLock);
			Utils.SetupAction(this, "Abort", AbortLock);
		
			myLockTime = Utils.SetupSliderFloat(this, "Lock Time (min)", 10, 1, 30, false);
			myLockTime.slider.wholeNumbers = true;
			myCountdownTime = Utils.SetupSliderFloat(this, "Countdown (sec)", 10, 5, 60, false);
			myCountdownTime.slider.wholeNumbers = true;
			
			myWarnRadius = Utils.SetupSliderFloat(this, "Warn Radius (cm)", 15, 5, 49, false);
			myWarnRadius.slider.wholeNumbers = true;
			myWarnRadius.setCallbackFunction += (float v) => { myBreachRadius.val = Mathf.Max(v+1, myBreachRadius.val);	};
			
			myWarnAngle = Utils.SetupSliderFloat(this, "Warn Angle (deg)", 10, 3, 59, false);
			myWarnAngle.slider.wholeNumbers = true;
			myWarnAngle.setCallbackFunction += (float v) => { myBreachAngle.val = Mathf.Max(v+1, myBreachAngle.val); };
			
			myBreachRadius = Utils.SetupSliderFloat(this, "Breach Radius (cm)", 30, 6, 50, false);	
			myBreachRadius.slider.wholeNumbers = true;			
			myBreachRadius.setCallbackFunction += (float v) => { myWarnRadius.val = Mathf.Min(v-1, myWarnRadius.val); };
			
			myBreachAngle = Utils.SetupSliderFloat(this, "Breach Angle (deg)", 15, 4, 60, false);
			myBreachAngle.slider.wholeNumbers = true;
			myBreachAngle.setCallbackFunction += (float v) => { myWarnAngle.val = Mathf.Min(v-1, myWarnAngle.val); };
			
			myBreachDuration = Utils.SetupSliderFloat(this, "Breach Duration (sec)", 15, 5, 60, false);
			myBreachDuration.slider.wholeNumbers = true;
			
			Utils.SetupButton(this, "Actions OnInit", myEventInit.OpenPanel, true);
			Utils.SetupButton(this, "Actions OnLock", myEventLock.OpenPanel, true);
			Utils.SetupButton(this, "Actions OnUnlock", myEventUnlock.OpenPanel, true);
			Utils.SetupButton(this, "Actions OnClear", myEventClear.OpenPanel, true);
			Utils.SetupButton(this, "Actions OnWarn", myEventWarn.OpenPanel, true);
			Utils.SetupButton(this, "Actions OnBreach", myEventBreach.OpenPanel, true);
			
			Utils.SetupSpacer(this, 135.0f, true);
		
			myHeadupAngleVR = Utils.SetupSliderFloat(this, "Head-up Display Position VR (deg)", 21, -45, 45, true);
			myHeadupAngleDesktop = Utils.SetupSliderFloat(this, "Head-up Display Position Desktop (deg)", 15, -30, 30, true);
			
			FindFocusPoint();
			sc.onAtomUIDsChangedHandlers += OnAtomUIDsChanged;
			sc.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}
		
		private void OnAtomUIDsChanged(List<string> atomUIDs)
		{
			FindFocusPoint();
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			myEventInit.SyncAtomNames();
			myEventLock.SyncAtomNames();
			myEventUnlock.SyncAtomNames();
			myEventClear.SyncAtomNames();
			myEventWarn.SyncAtomNames();
			myEventBreach.SyncAtomNames();
		}
		
		private void OnAssetBundleLoaded(Request aRequest)
		{	
			GameObject prefab = aRequest.assetBundle?.LoadAsset<GameObject>(BASE_PATH + "VirtualLockUI.prefab");
			
			if (prefab != null)
			{
				myCanvas = Instantiate(prefab).transform;
				myTimerPanel = myCanvas.Find("Timer")?.GetComponent<Image>();
				myMessagePanel = myCanvas.Find("Message")?.GetComponent<Image>();
				myHeadPanel = myCanvas.Find("Head")?.GetComponent<Image>();
				myHandPanelL = myCanvas?.Find("LeftHand")?.GetComponent<Image>();
				myHandPanelR = myCanvas?.Find("RightHand")?.GetComponent<Image>();
				
				
				
				myTimerText = myTimerPanel?.transform?.Find("Time")?.GetComponent<Text>();
				myMessageText = myMessagePanel?.transform?.Find("Label")?.GetComponent<Text>();
				myHeadDistanceText = myHeadPanel?.transform?.Find("Distance")?.GetComponent<Text>();
				myHeadAngleText = myHeadPanel?.transform?.Find("Angle")?.GetComponent<Text>();				
				myHandDistanceTextL = myHandPanelL?.transform?.Find("Distance")?.GetComponent<Text>();
				myHandDistanceTextR = myHandPanelR?.transform?.Find("Distance")?.GetComponent<Text>();
				myCanvas.gameObject.SetActive(false);
			}
		}
		
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myEventInit.Name] = myEventInit.GetJSON(base.subScenePrefix);
				jc[myEventLock.Name] = myEventLock.GetJSON(base.subScenePrefix);
				jc[myEventUnlock.Name] = myEventUnlock.GetJSON(base.subScenePrefix);
				jc[myEventClear.Name] = myEventClear.GetJSON(base.subScenePrefix);
				jc[myEventWarn.Name] = myEventWarn.GetJSON(base.subScenePrefix);
				jc[myEventBreach.Name] = myEventBreach.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myEventInit.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myEventLock.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myEventUnlock.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myEventClear.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myEventWarn.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myEventBreach.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}
			
		private void OnDisable()
		{
			FinishLock();
		}
		
		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDsChangedHandlers -= OnAtomUIDsChanged;
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			
			if (myCanvas != null)
				Destroy(myCanvas.gameObject);			
			if (myBundleURL != null)
                AssetLoader.DoneWithAssetBundleFromFile(myBundleURL);
			
			RemoveTrigger(myEventInit);
			RemoveTrigger(myEventLock);
			RemoveTrigger(myEventUnlock);
			RemoveTrigger(myEventClear);
			RemoveTrigger(myEventWarn);
			RemoveTrigger(myEventBreach);
		}
		
		private void RemoveTrigger(Trigger trigger)
		{
			if (trigger != null)
				trigger.Remove();
		}
					
		private void Update()
		{			
			if (!myEngaged || myCanvas == null)
				return;
						
			if (myCountdown > 0.0f)
			{
				// Init phase, counting down
				myCountdown -= Time.unscaledDeltaTime;
				if (myCountdown <= 0.0f)
					EnableLock();
				else
					UpdateTextCountdown();
			}
			else if (myState == STATE_BREACH)
			{
				// Breach phase
				myBreachTimer -= Time.unscaledDeltaTime;
				if (myBreachTimer <= 0.0f)
					DisableLock();
				else
					UpdateTextBreach();
			}
			else if (myTimer > 0.0f)
			{
				// Lock phase
				myTimer -= Time.unscaledDeltaTime;
				if (myTimer <= 0.0f)
				{
					DisableLock();
				}
				else if (HaveInput())
				{
					BreachedLock();
				}
				else
				{
					if (myFadeClock > 0.0f)
						myFadeClock -= Time.unscaledDeltaTime;
					myCanvas.gameObject.SetActive(myFadeClock >= 0.0f);
					
					UpdateTextTimer();
					UpdateTextDistance(myHeadDistanceText, ref myHeadDistance, myHeadPosition, myHeadControl);
					UpdateTextDistance(myHandDistanceTextL, ref myHandDistanceL, myHandPositionL, myHandControlL);
					UpdateTextDistance(myHandDistanceTextR, ref myHandDistanceR, myHandPositionR, myHandControlR);
					UpdateTextHeadAngle();				
					UpdateState();				
					myAlpha = Mathf.Clamp01(myFadeClock);
					UpdateColor(myHeadPanel, myHeadDistance);
					UpdateColor(myHandPanelL, myHandDistanceL);
					UpdateColor(myHandPanelR, myHandDistanceR);
					UpdateColorTimer();
				}
			}
			else 
			{
				// Unlock phase, show "Unlocked" message
				myTimer -= Time.unscaledDeltaTime;
				myAlpha = Mathf.Clamp01(myTimer + FADE_TIME);
				UpdateTextTimer();
				UpdateColorTimer();
				if (myTimer <= -FADE_TIME)
				{
					FinishLock();
					return;
				}
			}
			
			SuperController sc = SuperController.singleton;
			bool isVR = sc.isOVR || sc.isOpenVR;
			float angleDeg = isVR ? myHeadupAngleVR.val : myHeadupAngleDesktop.val;
			float angleRad = angleDeg * Mathf.Deg2Rad;
			Vector3 positionOffset = new Vector3(0.0f, Mathf.Sin(angleRad), Mathf.Cos(angleRad)) * HEADUP_DISTANCE;
			Quaternion rotationOffset = isVR ? Quaternion.Euler(-angleDeg, 0.0f, 0.0f) : Quaternion.identity;
			myCanvas.position = myHeadControl.TransformPoint(positionOffset);
			myCanvas.rotation = myHeadControl.rotation * rotationOffset;
		}
		
		private void UpdateTextDistance(Text text, ref int lastDistance, Vector3 storedPosition, Transform actualPosition)
		{
			if (text == null)
			{
				lastDistance = -1;
				return;
			}
				
			
			float d = Vector3.Distance(storedPosition, actualPosition.position);
			int currentDistance = Mathf.RoundToInt(d * 100.0f);
			if (lastDistance != currentDistance)
			{
				lastDistance = currentDistance;
				text.text = currentDistance + "cm";
			}
		}
		
		private void UpdateTextHeadAngle()
		{
			if (myHeadAngleText == null)
			{
				myHeadAngle = -1;
				return;
			}
			
			int currentAngle = 0;
			if (myFocusPoint != null)
			{
				Vector3 focusDir = myFocusPoint.position - myHeadControl.position;
				float focusDist = focusDir.magnitude;
				focusDir = (focusDist > 0.01f) ? focusDir/focusDist : myHeadControl.forward;
				float a = Vector3.Angle(myHeadControl.forward, focusDir);
				currentAngle = Mathf.RoundToInt(a);
			}
			
			if (myHeadAngle != currentAngle)
			{
				myHeadAngle = currentAngle;
				myHeadAngleText.text = currentAngle + "°";
			}
		}
		
		private void UpdateTextTimer()
		{
			int time = Mathf.Max(Mathf.CeilToInt(myTimer), 0);
			if (time == myTimerSeconds)
				return;
			
			myTimerSeconds = time;
			int seconds = time % 60;
			int minutes = time / 60;
			int hours = minutes / 60;
			minutes = minutes % 60;
			myTimerText.text = string.Format("{0:00}:{1:00}:{2:00}", hours, minutes, seconds);
		}
		
		private void UpdateTextCountdown()
		{
			int countdownSeconds = Mathf.CeilToInt(myCountdown);
			if (myCountdownSeconds != countdownSeconds)
			{
				myCountdownSeconds = countdownSeconds;
				myMessageText.text = string.Format("Locking in {0}", myCountdownSeconds);
			}
		}
		
		private void UpdateTextBreach()
		{
			int breachTimerSeconds = Mathf.CeilToInt(myBreachTimer);
			if (myBreachTimerSeconds != breachTimerSeconds)
			{
				myBreachTimerSeconds = breachTimerSeconds;
				myMessageText.text = string.Format("Lock breached! {0}", myBreachTimerSeconds);
			}
		}
		
		private void UpdateColor(Image panel, float distance)
		{
			float t = Mathf.InverseLerp(myWarnRadius.val, myBreachRadius.val, distance);		
			if (panel == myHeadPanel)
				t = Mathf.Max(t, Mathf.InverseLerp(myWarnAngle.val, myBreachAngle.val, myHeadAngle));
			
			t = myState < STATE_BREACH ? Mathf.Clamp01(t) : 1.0f;
			Color c = Color.LerpUnclamped(BASE_COLOR, BREACH_COLOR, t);
			c.a = myAlpha;
			panel.color = c;
		}
		
		private void UpdateColorTimer()
		{
			Color c = myState < STATE_BREACH ? BASE_COLOR : BREACH_COLOR;
			c.a = myAlpha;
			myTimerPanel.color = c;
			myMessagePanel.color = c;
		}
		
		private void UpdateState()
		{
			float maxDistance = Mathf.Max(myHeadDistance, Mathf.Max(myHandDistanceL, myHandDistanceR));
			bool outer = maxDistance >= myBreachRadius.val || myHeadAngle >= myBreachAngle.val;
			bool inner = maxDistance >= myWarnRadius.val || myHeadAngle >= myWarnAngle.val;
			if (outer)
			{
				if (myState < STATE_BREACH)
				{
					BreachedLock();
				}
				myFadeClock = FADE_TIME;
			}
			else if (inner)
			{
				if (myState < STATE_WARN)
				{
					myEventWarn.Trigger();
					myState = STATE_WARN;
				}
				myFadeClock = FADE_TIME;
			}
			else
			{
				if (myState != STATE_CLEAR && myState != STATE_BREACH)
				{
					myEventClear.Trigger();
					myState = STATE_CLEAR;
				}
			}
			
			if (myState == STATE_BREACH)
			{
				myFadeClock = FADE_TIME;
			}
		}
		
		private bool HaveInput()
		{
			SuperController sc = SuperController.singleton;
			return sc.GetLeftSelect()
				|| sc.GetMenuMoveLeft()
				|| sc.GetRightSelect()
				|| sc.GetMenuMoveRight()
				|| (sc.isOVR && !OVRManager.hasInputFocus);
		}
		
		private void InitLock()
		{
			if (myEngaged && myCountdown < 0.0f && myTimer > 0.0f)
			{
				SuperController.LogError("VirtualLock can't re-init while locked! That's kind of the point...");
				return;
			}
						
			myBreachTimer = myBreachDuration.val;
			myCountdown = myCountdownTime.val;
			myTimer = myLockTime.val * 60.0f;
			myCountdownSeconds = -1;
			myTimerSeconds = -1;
			myBreachTimerSeconds = -1;
			myState = STATE_CLEAR;
			myEngaged = true;
			
			UpdateTextTimer();
			
			myTimerPanel.color = BASE_COLOR;
			myMessagePanel.color = BASE_COLOR;
			myMessagePanel.gameObject.SetActive(true);
			myHeadPanel.gameObject.SetActive(false);
			myHandPanelL.gameObject.SetActive(false);
			myHandPanelR.gameObject.SetActive(false);
			
			myCanvas.gameObject.SetActive(true);
			
			
			myEventInit.Trigger();
		}
		
		private void AbortLock()
		{	
			if (myCountdown < 0.0f && myTimer > 0.0f)
			{		
				SuperController.LogError("VirtualLock can't abort while locked! That's kind of the point...");
				return;
			}
			
			DisableLock();
		}
		
		private void BreachedLock()
		{
			myTimerPanel.color = BREACH_COLOR;
			myMessagePanel.color = BREACH_COLOR;
			myBreachTimerSeconds = -1;
			UpdateTextBreach();
			myMessagePanel.gameObject.SetActive(true);
			myHeadPanel.gameObject.SetActive(false);
			myHandPanelL.gameObject.SetActive(false);
			myHandPanelR.gameObject.SetActive(false);
			myState = STATE_BREACH;
			
			myEventBreach.Trigger();
		}
		
		private void EnableLock()
		{		
			myHeadPosition = myHeadControl.position;
			myHandPositionL = myHandControlL.position;
			myHandPositionR = myHandControlR.position;
			myHeadDistance = -1;
			myHandDistanceL = -1;
			myHandDistanceR = -1;
			
			myCountdown = -1.0f;
			myFadeClock = FADE_TIME;
			
			myMessagePanel.gameObject.SetActive(false);
			myHeadPanel.gameObject.SetActive(true);
			myHandPanelL.gameObject.SetActive(true);
			myHandPanelR.gameObject.SetActive(true);
			
			myEventLock.Trigger();
		}
		
		private void DisableLock()
		{	
			myCountdown = -1.0f;
			myTimer = 0.0f;
		
			myState = STATE_CLEAR;
			myMessageText.text = "Unlocked";
			myMessagePanel.gameObject.SetActive(true);
			myHeadPanel.gameObject.SetActive(false);
			myHandPanelL.gameObject.SetActive(false);
			myHandPanelR.gameObject.SetActive(false);
			
			myCanvas.gameObject.SetActive(true);
			
			myEventUnlock.Trigger();
		}
		
		private void FinishLock()
		{
			if (myEngaged && myCountdown < 0.0f && myTimer > 0.0f)
				BreachedLock(); // plugin disabled while in lock!
			
			if (myCanvas != null)
				myCanvas.gameObject.SetActive(false);
			myEngaged = false;
		}
		
		private void FindFocusPoint()
		{
			Atom atom = GetAtomById("VirtualLockFocusPoint");
			myFocusPoint = atom?.GetStorableByID("control")?.transform;
		}
    }
}
